@@ -66,15 +66,21 @@ import sys
 import types
 import os
 import gc
-import collections
 import decimal
 import datetime
 import uuid
 import array
 import math
+import numbers
 import codecs
+import collections
+if sys.version_info >= (3, 4):
+    from collections.abc import KeysView, ValuesView, ItemsView
+else:
+    from collections import KeysView, ValuesView, ItemsView
 
-__version__ = "1.19"
+
+__version__ = "1.24"
 __all__ = ["dump", "dumps", "load", "loads", "register_class", "unregister_class", "tobytes"]
 
 can_use_set_literals = sys.version_info >= (3, 2)  # check if we can use set literals
@@ -145,18 +151,25 @@ def _ser_DictView(obj, serializer, outputstream, indentlevel):
     serializer.ser_builtins_list(obj, outputstream, indentlevel)
 
 
-_special_classes_registry = {
-    collections.KeysView: _ser_DictView,
-    collections.ValuesView: _ser_DictView,
-    collections.ItemsView: _ser_DictView
-}
-if sys.version_info >= (2, 7):
-    _special_classes_registry[collections.OrderedDict] = _ser_OrderedDict
-if sys.version_info >= (3, 4):
-    import enum
-    def _ser_Enum(obj, serializer, outputstream, indentlevel):
-        serializer._serialize(obj.value, outputstream, indentlevel)
-    _special_classes_registry[enum.Enum] = _ser_Enum
+_special_classes_registry = collections.OrderedDict()   # must be insert-order preserving to make sure of proper precedence rules
+
+
+def _reset_special_classes_registry():
+    _special_classes_registry.clear()
+    _special_classes_registry[KeysView] = _ser_DictView
+    _special_classes_registry[ValuesView] = _ser_DictView
+    _special_classes_registry[ItemsView] = _ser_DictView
+    if sys.version_info >= (2, 7):
+        _special_classes_registry[collections.OrderedDict] = _ser_OrderedDict
+    if sys.version_info >= (3, 4):
+        import enum
+
+        def _ser_Enum(obj, serializer, outputstream, indentlevel):
+            serializer._serialize(obj.value, outputstream, indentlevel)
+        _special_classes_registry[enum.Enum] = _ser_Enum
+
+
+_reset_special_classes_registry()
 
 
 def unregister_class(clazz):
@@ -278,7 +291,8 @@ class Serializer(object):
         """
         Initialize the serializer.
         indent=indent the output over multiple lines (default=false)
-        setLiterals=use set-literals or not (set to False if you need compatibility with Python < 3.2). Serpent chooses a sensible default for you.
+        setLiterals=use set-literals or not (set to False if you need compatibility with Python < 3.2).
+        Serpent chooses a sensible default for you.
         module_in_classname = include module prefix for class names or only use the class name itself
         """
         self.indent = indent
@@ -286,7 +300,7 @@ class Serializer(object):
         self.module_in_classname = module_in_classname
         self.serialized_obj_ids = set()
         self.special_classes_registry_copy = None
-        self.maximum_level = min(sys.getrecursionlimit() // 3.5, 1000)
+        self.maximum_level = min(sys.getrecursionlimit() // 5, 1000)
 
     def serialize(self, obj):
         """Serialize the object tree to bytes."""
@@ -295,7 +309,7 @@ class Serializer(object):
         if self.set_literals:
             header += "python3.2\n"   # set-literals require python 3.2+ to deserialize (ast.literal_eval limitation)
         else:
-            header += "python2.6\n"   # don't change this even though we don't support 2.6 any longer, otherwise we can't read older serpent strings
+            header += "python2.6\n"   # don't change this, otherwise we can't read older serpent strings
         out = [header]
         if os.name == "java" and type(obj) is buffer:
             obj = bytearray(obj)
@@ -314,7 +328,8 @@ class Serializer(object):
 
     def _serialize(self, obj, out, level):
         if level > self.maximum_level:
-            raise ValueError("Object graph nesting too deep. Increase serializer.maximum_level if you think you need more.")
+            raise ValueError("Object graph nesting too deep. Increase serializer.maximum_level if you think you need more, "
+                             " but this may cause a RecursionError instead if Python's recursion limit doesn't allow it.")
         t = type(obj)
         if t in _translate_types:
             obj = _translate_types[t](obj)
@@ -440,6 +455,17 @@ class Serializer(object):
         self.serialized_obj_ids.discard(id(list_obj))
     dispatch[list] = ser_builtins_list
 
+    def _check_hashable_type(self, t):
+        if t not in (bool, bytes, str, tuple) and not issubclass(t, numbers.Number):
+            if sys.version_info >= (3, 4):
+                import enum
+                if issubclass(t, enum.Enum):
+                    return
+            elif sys.version_info < (3, 0) and t is unicode:
+                return
+            raise TypeError("one of the keys in a dict or set is not of a primitive hashable type: " +
+                            str(t) + ". Use simple types as keys or use a list or tuple as container.")
+
     def ser_builtins_dict(self, dict_obj, out, level):
         if id(dict_obj) in self.serialized_obj_ids:
             raise ValueError("Circular reference detected (dict)")
@@ -457,6 +483,7 @@ class Serializer(object):
                 sorted_items = dict_items
             for key, value in sorted_items:
                 append(indent_chars_inside)
+                self._check_hashable_type(type(key))
                 serialize(key, out, level + 1)
                 append(": ")
                 serialize(value, out, level + 1)
@@ -466,6 +493,7 @@ class Serializer(object):
         else:
             append("{")
             for key, value in dict_obj.items():
+                self._check_hashable_type(type(key))
                 serialize(key, out, level + 1)
                 append(":")
                 serialize(value, out, level + 1)
@@ -494,6 +522,7 @@ class Serializer(object):
                 sorted_elts = set_obj
             for elt in sorted_elts:
                 append(indent_chars_inside)
+                self._check_hashable_type(type(elt))
                 serialize(elt, out, level + 1)
                 append(",\n")
             del out[-1]  # remove the last ,\n
@@ -501,6 +530,7 @@ class Serializer(object):
         elif set_obj:
             append("{")
             for elt in set_obj:
+                self._check_hashable_type(type(elt))
                 serialize(elt, out, level + 1)
                 append(",")
             del out[-1]  # remove the last ,
@@ -522,11 +552,11 @@ class Serializer(object):
     def ser_datetime_datetime(self, datetime_obj, out, level):
         out.append(repr(datetime_obj.isoformat()))
     dispatch[datetime.datetime] = ser_datetime_datetime
-    
+
     def ser_datetime_date(self, date_obj, out, level):
         out.append(repr(date_obj.isoformat()))
     dispatch[datetime.date] = ser_datetime_date
-    
+
     if os.name == "java" or sys.version_info < (2, 7):    # jython bug http://bugs.jython.org/issue2010
         def ser_datetime_timedelta(self, timedelta_obj, out, level):
             secs = ((timedelta_obj.days * 86400 + timedelta_obj.seconds) * 10 ** 6 + timedelta_obj.microseconds) / 10 ** 6
@@ -592,7 +622,8 @@ class Serializer(object):
                             value[slot] = getattr(obj, slot)
                         value["__class__"] = self.get_class_name(obj)
                     else:
-                        raise TypeError("don't know how to serialize class " + str(obj.__class__) + ". Give it vars() or an appropriate __getstate__")
+                        raise TypeError("don't know how to serialize class " +
+                                        str(obj.__class__) + ". Give it vars() or an appropriate __getstate__")
             self._serialize(value, out, level)
         finally:
             self.serialized_obj_ids.discard(id(obj))
